import time
from json import dumps
from typing import List

import pandas as pd
from tqdm import tqdm

from ..helper import format_result
from .project import get_project


def get_asset(client, asset_id: str):
    result = client.execute('''
    query {
      getAsset(assetID: "%s") {
        id
        externalId
        content
        filename
        isInstructions
        instructions
        isHoneypot
        consensusMark
        honeypotMark
        status
        isUsedForConsensus
        labels {
          id
          createdAt
          secondsToLabel
          totalSecondsToLabel
          labelType
          jsonResponse
          isLatestLabelForUser
          author {
            id
          }
        }
        locks {
          id
          author {
            email
          }
          createdAt
          lockType
        }
      }
    }
    ''' % (asset_id))
    return format_result('getAsset', result)


def get_assets(client, project_id: str, skip: int = None, first: int = None,
               external_id_contains: str = None,
               status_in: List[str] = None,
               author_in: List[str] = None,
               consensus_mark_gt: float = None,
               consensus_mark_lt: float = None, honeypot_mark_gt: float = None,
               honeypot_mark_lt: float = None, skipped: bool = None, format: str = None,
               disable_tqdm: bool = False):
    formatted_skip = 0 if skip is None else skip
    formatted_first = 100
    formatted_external_id_contains = 'null' if external_id_contains is None else f'"{external_id_contains}"'
    formatted_status_in = dumps([]) if status_in is None else dumps(status_in)
    formatted_author_in = dumps([]) if author_in is None else dumps(author_in)
    formatted_consensus_mark_gt = 'null' if consensus_mark_gt is None else f'{consensus_mark_gt}'
    formatted_consensus_mark_lt = 'null' if consensus_mark_lt is None else f'{consensus_mark_lt}'
    formatted_honeypot_mark_gt = 'null' if honeypot_mark_gt is None else f'{honeypot_mark_gt}'
    formatted_honeypot_mark_lt = 'null' if honeypot_mark_lt is None else f'{honeypot_mark_lt}'
    formatted_skipped = 'null' if skipped is None else f'{skipped}'.lower(
    )
    project = get_project(client, project_id)
    number_of_assets = project['numberOfAssets']
    total = number_of_assets if first is None else first
    with tqdm(total=total, disable=disable_tqdm) as pbar:
        paged_assets = []
        while True:
            result = client.execute('''
            query {
              getAssetsWithSearch(projectID: "%s", skip: %d, first: %d
                where: {
                  externalIdContains: %s
                  statusIn: %s
                  authorIn: %s
                  consensusMarkGt: %s
                  consensusMarkLt: %s
                  honeypotMarkGt: %s
                  honeypotMarkLt: %s
                  skipped: %s
                }) {
                id
                externalId
                content
                filename
                isInstructions
                instructions
                isHoneypot
                consensusMark
                honeypotMark
                calculatedConsensusMark
                calculatedHoneypotMark
                status
                isUsedForConsensus
                jsonMetadata
                priority
                labels {
                  id
                  createdAt
                  labelType
                  jsonResponse
                  isLatestLabelForUser
                  skipped
                  author {
                    id
                    email
                  }
                }
              }
            }
            ''' % (project_id, formatted_skip, formatted_first, formatted_external_id_contains,
                   formatted_status_in,
                   formatted_author_in,
                   formatted_consensus_mark_gt,
                   formatted_consensus_mark_lt,
                   formatted_honeypot_mark_gt,
                   formatted_honeypot_mark_lt,
                   formatted_skipped))
            assets = format_result('getAssetsWithSearch', result)
            if assets is None or (first is not None and len(paged_assets) == first):
                if format == 'pandas':
                    return pd.DataFrame(paged_assets)
                return paged_assets
            if first is not None:
                assets = assets[:max(0, first - len(paged_assets))]
            paged_assets += assets
            formatted_skip += formatted_first
            pbar.update(len(assets))


def export_assets(**kwargs):
    print('export_assets has been deprecated. Please use get_assets instead.')


def get_assets_with_search(**kwargs):
    print('get_assets_with_search has been renamed in get_assets. Please use get_assets instead.')


def get_assets_by_external_id(client, project_id: str, external_id: str):
    result = client.execute('''
    query {
      getAssetsByExternalId(projectID: "%s", externalID: "%s") {
          id
          content
          externalId
          createdAt
          updatedAt
          isHoneypot
          isUsedForConsensus
          status
          labels {
            author {
              id
              email
            }
            labelType
            jsonResponse
            createdAt
            secondsToLabel
            totalSecondsToLabel
            honeypotMark
          }
      }
    }
    ''' % (project_id, external_id))
    return format_result('getAssetsByExternalId', result)


def get_next_asset_from_label(client, label_asset_id: str, want_instructions_only: bool):
    result = client.execute('''
    query {
      getNextAssetFromLabel(labelAssetID: "%s", wantInstructionsOnly: %s, where: {}) {
        id
      }
    }
    ''' % (label_asset_id, str(want_instructions_only).lower()))
    return format_result('getNextAssetFromLabel', result)


def get_next_asset_from_project(client, project_id: str, want_instructions_only: bool):
    result = client.execute('''
    query {
      getNextAssetFromProject(projectID: "%s", wantInstructionsOnly: %s) {
        id
      }
    }
    ''' % (project_id, str(want_instructions_only).lower()))
    return format_result('getNextAssetFromProject', result)


def export_assets(client, project_id: str):
    result = client.execute('''
    query {
      exportAssets(projectID: "%s") {
        id
        content
        externalId
        createdAt
        updatedAt
        isHoneypot
        isInstructions
        status
        labels {
          id
          author {
            id
            email
          }
          labelType
          jsonResponse
          createdAt
          secondsToLabel
          totalSecondsToLabel
          honeypotMark
          isLatestLabelForUser
        }
      }
    }
    ''' % (project_id))
    return format_result('exportAssets', result)



from ..helper import format_result


def signin(client, email: str, password: str):
    result = client.execute('''
    mutation {
      signIn(email: "%s", password: "%s") {
        id
        token
        user {
          id
        }
      }
    }
    ''' % (email, password))
    return format_result('signIn', result)


def create_user(client, name: str, email: str, password: str, phone: str, organization_role: str):
    result = client.execute('''
    mutation {
      createUser(name: "%s",
      email: "%s",
      password: "%s",
      phone: "%s",
      organizationRole: %s) {
        id
      }
    }
    ''' % (name, email, password, phone, organization_role))
    return format_result('createUser', result)


def create_user_from_email_if_not_exists(client, name: str, email: str, organization_role: str, project_id: str):
    result = client.execute('''
    mutation {
      createUserFromEmailIfNotExists(name: "%s",
      email: "%s",
      organizationRole: %s,
      projectID: "%s") {
        id
      }
    }
    ''' % (name, email, organization_role, project_id))
    return format_result('createUserFromEmailIfNotExists', result)


def update_user(client, user_id: str, name: str, email: str, phone: str, organization_role: str):
    result = client.execute('''
    mutation {
      updateUser(userID: "%s",
      name: "%s",
      email: "%s",
      phone: "%s",
      organizationRole: %s) {
        id
      }
    }
    ''' % (user_id, name, email, phone, organization_role))
    return format_result('updateUser', result)


def update_password(client, email: str, old_password: str, new_password_1: str, new_password_2: str):
    result = client.execute('''
    mutation {
      updatePassword(email: "%s",
      oldPassword: "%s",
      newPassword1: "%s",
      newPassword2: "%s") {
        id
      }
    }
    ''' % (email, old_password, new_password_1, new_password_2))
    return format_result('updatePassword', result)


def reset_password(client, email: str):
    result = client.execute('''
    mutation {
      resetPassword(email: "%s") {
        id
      }
    }
    ''' % email)
    return format_result('resetPassword', result)



import json
import time

from six.moves import urllib


class GraphQLClient:
    def __init__(self, endpoint, session=None):
        self.endpoint = endpoint
        self.session = session
        self.token = None
        self.headername = None

    def execute(self, query, variables=None):
        return self._send(query, variables)

    def inject_token(self, token, headername='Authorization'):
        self.token = token
        self.headername = headername

    def _send(self, query, variables):
        data = {'query': query,
                'variables': variables}
        headers = {'Accept': 'application/json',
                   'Content-Type': 'application/json'}

        if self.token is not None:
            headers[self.headername] = '{}'.format(self.token)

        if self.session is not None:
            try:
                number_of_trials = 3
                for n in range(number_of_trials):
                    req = self.session.post(self.endpoint, json.dumps(
                        data).encode('utf-8'), headers=headers)
                    if req.status_code == 200:
                        break
                    time.sleep(1)
                return req.json()
            except Exception as e:
                print('Request failed with error:\n')
                if req:
                    print(req.content)
                print('')
                raise e

        req = urllib.request.Request(
            self.endpoint, json.dumps(data).encode('utf-8'), headers)
        try:
            response = urllib.request.urlopen(req)
            str_json = response.read().decode('utf-8')
            return json.loads(str_json)
        except urllib.error.HTTPError as e:
            print((e.read()))
            print('')
            raise e



from ..helper import format_result, json_escape


def create_honeypot(client, asset_id: str, json_response: str):
    result = client.execute('''
    mutation {
      createHoneypot(
        assetID: "%s",
        jsonResponse: "%s") {
          id
          author {
            id
            email
          }
          labelType
          jsonResponse
          createdAt
          secondsToLabel
          totalSecondsToLabel
          honeypotMark
      }
    }
    ''' % (asset_id, json_escape(json_response)))
    return format_result('createHoneypot', result)



from ..helper import format_result


def get_tools(client, project_id: str):
    result = client.execute('''
    query {
      getTools(projectID: "%s") {
        id
        name
        toolType
        jsonSettings
      }
    }
    ''' % (project_id))
    return format_result('getTools', result)



from ..helper import format_result, json_escape


def update_tool(client, tool_id: str, project_id: str, name: str, tool_type: str, json_settings: str):
    result = client.execute('''
    mutation {
      updateTool(toolID: "%s",
        projectID: "%s",
        name: "%s",
        toolType: %s,
        jsonSettings: "%s") {
          id
      }
    }
    ''' % (tool_id, project_id, name, tool_type, json_escape(json_settings)))
    return format_result('updateTool', result)


def append_to_tools(client, project_id: str, name: str, tool_type: str, json_settings: str):
    result = client.execute('''
    mutation {
      appendToTools(
        projectID: "%s",
        name: "%s",
        toolType: %s,
        jsonSettings: "%s") {
          id
      }
    }
    ''' % (project_id, name, tool_type, json_escape(json_settings)))
    return format_result('appendToTools', result)


def delete_from_tools(client, tool_id: str):
    result = client.execute('''
    mutation {
      deleteFromTools(toolID: "%s") {
        id
      }
    }
    ''' % (tool_id))
    return format_result('deleteFromTools', result)



import kili.mutations.asset
import kili.mutations.label
import kili.mutations.lock
import kili.mutations.organization
import kili.mutations.project
import kili.mutations.tool
import kili.mutations.user
import kili.quality.consensus
import kili.quality.honeypot
import kili.queries.asset
import kili.queries.label
import kili.queries.lock
import kili.queries.project
import kili.queries.tool
import kili.queries.user


class Playground(object):

    def __init__(self, auth=None):
        """Create an instance of KiliPlayground."""
        self.auth = auth

    # Mutations Asset

    def create_assets(self, **kwargs):
        return kili.mutations.asset.create_assets(self.auth.client, **kwargs)

    def delete_assets_by_external_id(self, **kwargs):
        return kili.mutations.asset.delete_assets_by_external_id(self.auth.client, **kwargs)

    def append_to_dataset(self, **kwargs):
        return kili.mutations.asset.append_to_dataset(self.auth.client, **kwargs)

    def append_many_to_dataset(self, **kwargs):
        return kili.mutations.asset.append_many_to_dataset(self.auth.client, **kwargs)

    def update_asset(self, **kwargs):
        return kili.mutations.asset.update_asset(self.auth.client, **kwargs)

    def update_properties_in_asset(self, **kwargs):
        return kili.mutations.asset.update_properties_in_asset(self.auth.client, **kwargs)

    def delete_from_dataset(self, **kwargs):
        return kili.mutations.asset.delete_from_dataset(self.auth.client, **kwargs)

    def delete_many_from_dataset(self, **kwargs):
        return kili.mutations.asset.delete_many_from_dataset(self.auth.client, **kwargs)

    def force_update_status(self, **kwargs):
        return kili.mutations.asset.force_update_status(self.auth.client, **kwargs)

    # Mutations Label

    def create_prediction(self, **kwargs):
        return kili.mutations.label.create_prediction(self.auth.client, **kwargs)

    def create_predictions(self, **kwargs):
        return kili.mutations.label.create_predictions(self.auth.client, **kwargs)

    def append_to_labels(self, **kwargs):
        return kili.mutations.label.append_to_labels(self.auth.client, **kwargs)

    def update_label(self, **kwargs):
        return kili.mutations.label.update_label(self.auth.client, **kwargs)

    def update_properties_in_label(self, **kwargs):
        return kili.mutations.label.update_properties_in_label(self.auth.client, **kwargs)

    # Mutations Lock

    def delete_locks(self, **kwargs):
        return kili.mutations.lock.delete_locks(self.auth.client, **kwargs)

    # Mutations Organization

    def create_organization(self, **kwargs):
        return kili.mutations.organization.create_organization(self.auth.client, **kwargs)

    def update_organization(self, **kwargs):
        return kili.mutations.organization.update_organization(self.auth.client, **kwargs)

    def delete_organization(self, **kwargs):
        return kili.mutations.organization.delete_organization(self.auth.client, **kwargs)

    # Mutations Project

    def create_project(self, **kwargs):
        return kili.mutations.project.create_project(self.auth.client, **kwargs)

    def delete_project(self, **kwargs):
        return kili.mutations.project.delete_project(self.auth.client, **kwargs)

    def append_to_roles(self, **kwargs):
        return kili.mutations.project.append_to_roles(self.auth.client, **kwargs)

    def update_properties_in_project(self, **kwargs):
        return kili.mutations.project.update_properties_in_project(self.auth.client, **kwargs)

    def update_interface_in_project(self, **kwargs):
        return kili.mutations.project.update_interface_in_project(self.auth.client, **kwargs)

    def create_empty_project(self, **kwargs):
        return kili.mutations.project.create_empty_project(self.auth.client, **kwargs)

    def update_project(self, **kwargs):
        return kili.mutations.project.update_project(self.auth.client, **kwargs)

    def update_role(self, **kwargs):
        return kili.mutations.project.update_role(self.auth.client, **kwargs)

    def delete_from_roles(self, **kwargs):
        return kili.mutations.project.delete_from_roles(self.auth.client, **kwargs)

    def update_properties_in_project_user(self, **kwargs):
        return kili.mutations.project.update_properties_in_project_user(self.auth.client, **kwargs)

    def force_project_kpis(self, **kwargs):
        return kili.mutations.project.force_project_kpis(self.auth.client, **kwargs)

    # Mutations Tool

    def update_tool(self, **kwargs):
        return kili.mutations.tool.update_tool(self.auth.client, **kwargs)

    def append_to_tools(self, **kwargs):
        return kili.mutations.tool.append_to_tools(self.auth.client, **kwargs)

    def delete_from_tools(self, **kwargs):
        return kili.mutations.tool.delete_from_tools(self.auth.client, **kwargs)

    # Mutations User

    def create_user(self, **kwargs):
        return kili.mutations.user.create_user(self.auth.client, **kwargs)

    def create_user_from_email_if_not_exists(self, **kwargs):
        return kili.mutations.user.create_user_from_email_if_not_exists(self.auth.client, **kwargs)

    def update_user(self, **kwargs):
        return kili.mutations.user.update_user(self.auth.client, **kwargs)

    def update_password(self, **kwargs):
        return kili.mutations.user.update_password(self.auth.client, **kwargs)

    def reset_password(self, **kwargs):
        return kili.mutations.user.reset_password(self.auth.client, **kwargs)

    # Quality Consensus

    def update_consensus_in_many_assets(self, **kwargs):
        return kili.quality.consensus.update_consensus_in_many_assets(self.auth.client, **kwargs)

    def force_consensus_for_project(self, **kwargs):
        return kili.quality.consensus.force_consensus_for_project(self.auth.client, **kwargs)

    # Quality Honeypot

    def create_honeypot(self, **kwargs):
        return kili.quality.honeypot.create_honeypot(self.auth.client, **kwargs)

    # Queries Asset

    def get_asset(self, **kwargs):
        return kili.queries.asset.get_asset(self.auth.client, **kwargs)

    def get_assets(self, **kwargs):
        return kili.queries.asset.get_assets(self.auth.client, **kwargs)

    def get_assets_with_search(self, **kwargs):
        return kili.queries.asset.get_assets_with_search(self.auth.client, **kwargs)

    def get_assets_by_external_id(self, **kwargs):
        return kili.queries.asset.get_assets_by_external_id(self.auth.client, **kwargs)

    def get_next_asset_from_label(self, **kwargs):
        return kili.queries.asset.get_next_asset_from_label(self.auth.client, **kwargs)

    def get_next_asset_from_project(self, **kwargs):
        return kili.queries.asset.get_next_asset_from_project(self.auth.client, **kwargs)

    # Queries Label

    def get_label(self, **kwargs):
        return kili.queries.label.get_label(self.auth.client, **kwargs)

    def get_latest_labels_for_user(self, **kwargs):
        return kili.queries.label.get_latest_labels_for_user(self.auth.client, **kwargs)

    def get_latest_labels(self, **kwargs):
        return kili.queries.label.get_latest_labels(self.auth.client, **kwargs)

    def export_labels_as_df(self, **kwargs):
        return kili.queries.label.export_labels_as_df(self.auth.client, **kwargs)

    # Queries Lock

    def get_locks(self, **kwargs):
        return kili.queries.lock.get_locks(self.auth.client, **kwargs)

    # Queries Project

    def get_projects(self, **kwargs):
        return kili.queries.project.get_projects(self.auth.client, **kwargs)

    def get_project(self, **kwargs):
        return kili.queries.project.get_project(self.auth.client, **kwargs)

    # Queries Tool

    def get_tools(self, **kwargs):
        return kili.queries.tool.get_tools(self.auth.client, **kwargs)

    # Queries User

    def get_user(self, **kwargs):
        return kili.queries.user.get_user(self.auth.client, **kwargs)


if __name__ == '__main__':
    """ Example of usage """
    from kili.authentication import KiliAuth
    from kili.playground import Playground
    kauth = KiliAuth()
    playground = Playground(kauth)
    assets = playground.export_assets(project_id="first-project")
    print(assets)



from ..helper import format_result


def create_organization(client, name: str, address: str, zip_code: str, city: str, country: str):
    result = client.execute('''
    mutation {
      createOrganization(name: "%s",
      address: "%s",
      zipCode: "%s",
      city: "%s",
      country: "%s") {
        id
      }
    }
    ''' % (name, address, zip_code, city, country))
    return format_result('createOrganization', result)


def update_organization(client, organization_id: str, name: str, address: str, zip_code: str, city: str, country: str):
    result = client.execute('''
    mutation {
      updateOrganization(organizationID: "%s",
      name: "%s",
      address: "%s",
      zipCode: "%s",
      city: "%s",
      country: "%s") {
        id
      }
    }
    ''' % (organization_id, name, address, zip_code, city, country))
    return format_result('updateOrganization', result)


def delete_organization(client, organization_id: str):
    result = client.execute('''
    mutation {
      deleteOrganization(organizationID: "%s") {
        id
      }
    }
    ''' % (organization_id))
    return format_result('deleteOrganization', result)



from json import dumps
from typing import List

from ..helper import format_result, json_escape


def create_prediction(client, asset_id: str, json_response: str):
    print('create_prediction is deprecated. Please use create_predictions instead. For an example, see: https://github.com/kili-technology/kili-playground/blob/master/recipes/import_predictions.py')


def create_predictions(client, project_id: str, external_id_array: List[str], json_response_array: List[str]):
    result = client.execute('''
    mutation {
      createPredictions(
        projectID: "%s",
        externalIDArray: %s,
        jsonResponseArray: %s) {
          id
      }
    }
    ''' % (project_id, dumps(external_id_array), dumps([dumps(elem) for elem in json_response_array])))
    return format_result('createPredictions', result)


def append_to_labels(client, author_id: str, json_response: str, label_asset_id: str, label_type: str, seconds_to_label: int, skipped: bool = False):
    result = client.execute('''
    mutation {
      appendToLabels(
        authorID: "%s",
        jsonResponse: "%s",
        labelAssetID: "%s",
        labelType: %s,
        secondsToLabel: %d,
        skipped: %s) {
          id
      }
    }
    ''' % (author_id, json_escape(json_response), label_asset_id, label_type, seconds_to_label, str(skipped).lower()))
    return format_result('appendToLabels', result)


def update_label(client, label_id: str, label_asset_id: str, review_asset_id: str, author_id: str, label_type: str, json_response: str, seconds_to_label: int):
    result = client.execute('''
    mutation {
      updateLabel(
        labelID: "%s",
        labelAssetID: "%s",
        reviewAssetID: "%s",
        authorID: "%s",
        labelType: %s,
        jsonResponse: "%s",
        secondsToLabel: %d) {
          id
      }
    }
    ''' % (label_id, label_asset_id, review_asset_id, author_id, label_type, json_escape(json_response), seconds_to_label))
    return format_result('updateLabel', result)


def update_properties_in_label(client, label_id: str, seconds_to_label: int = None, json_response: str = None):
    formatted_seconds_to_label = 'null' if seconds_to_label is None else f'{seconds_to_label}'
    formatted_json_response = 'null' if json_response is None else f'{json_response}'

    result = client.execute('''
        mutation {
          updatePropertiesInLabel(
            where: {id: "%s"},
            data: {
              secondsToLabel: %s
              jsonResponse: "%s"
            }
          ) {
            id
          }
        }
        ''' % (label_id, formatted_seconds_to_label, json_escape(json_response)))
    return format_result('updatePropertiesInLabel', result)



import json
from typing import List

import numpy as np
from shapely.geometry import Point
from shapely.geometry.polygon import Polygon
from .inter_rater import fleiss_kappa
from difflib import SequenceMatcher

from ..helper import format_result
from ..queries.asset import get_assets
from ..queries.tool import get_tools
from ..queries.project import get_project


def update_consensus_in_many_assets(client, asset_ids: List[str], consensus_marks: List[float], are_used_for_consensus: List[float]):
    asset_ids_in_string = '", "'.join(asset_ids)
    are_used_for_consensus_in_string = ', '.join(
        [str(is_used_for_consensus).lower() for is_used_for_consensus in are_used_for_consensus])
    result = client.execute('''
        mutation {
          updateConsensusInManyAssets(
            assetIDs: ["%s"],
            consensusMarks: %s
            areUsedForConsensus: [%s]
          ) {
              id
              consensusMark
              isUsedForConsensus
          }
        }
        ''' % (asset_ids_in_string, consensus_marks, are_used_for_consensus_in_string))
    return format_result('updateConsensusInManyAssets', result)


def compute_authors(labels):
    authors = []
    for label in labels:
        if label["labelType"] == "DEFAULT":
            authors.append(label["author"]["id"])
    return list(set(authors))


def compute_present_categories(asset):
    present_categories = []
    labels = asset["labels"]
    for label in labels:
        if label["labelType"] == "DEFAULT" and label["isLatestLabelForUser"]:
            response = json.loads(label["jsonResponse"])
            if "categories" not in response:
                continue
            categories = response["categories"]
            for checked_category in categories:
                present_categories.append(checked_category["name"])
    return list(set(present_categories))


def compute_bounding_polygons(labels, authors, grid_definition=100):
    all_bounding_poly = {}
    categories = []
    for author in authors:
        all_bounding_poly[author] = []
    for label in labels:
        if label["labelType"] == "DEFAULT" and label["isLatestLabelForUser"]:
            author = label["author"]["id"]
            annotations = json.loads(label["jsonResponse"])
            for annotation in annotations["annotations"]:
                multi_categories = annotation["description"][0]
                polygon_borders = []
                for border in annotation["boundingPoly"][0]["normalizedVertices"]:
                    polygon_borders.append(
                        (border['x'] * grid_definition, border['y'] * grid_definition))
                for category in multi_categories:
                    categories.append(category)
                    polygon = {"polygon": Polygon(
                        polygon_borders), "category": category}
                    all_bounding_poly[author].append(polygon)
    return all_bounding_poly, list(set(categories))


def compute_pixel_matrices_by_category(all_bounding_poly, categories, authors, grid_definition=100):
    nb_users = len(authors)
    kappa_matrices_by_category = {}
    for category in categories:
        kappa_matrices_by_category[category] = np.zeros(
            shape=(grid_definition ** 2, 2))
    for i in range(grid_definition):
        for j in range(grid_definition):
            nb_classification_for_pixel = {}
            for category in categories:
                nb_classification_for_pixel[category] = 0
            for author in authors:
                pixel_is_in_category = {}
                for category in categories:
                    pixel_is_in_category[category] = False
                for annotation in all_bounding_poly[author]:
                    polygon = annotation["polygon"]
                    if polygon.contains(Point(i, j)):
                        category = annotation["category"]
                        pixel_is_in_category[category] = True

                for category in categories:
                    if pixel_is_in_category[category]:
                        kappa_matrices_by_category[category][i *
                                                             grid_definition + j, 1] += 1
                        nb_classification_for_pixel[category] += 1

            for category in categories:
                kappa_matrices_by_category[category][i * grid_definition + j, 0] = nb_users - \
                    nb_classification_for_pixel[category]

    return kappa_matrices_by_category


def compute_consensus_for_assets(assets_for_consensus):
    consensus_by_asset = {}
    for asset in assets_for_consensus:
        categories = compute_present_categories(asset)
        dic_categories = {}
        for category in categories:
            dic_categories[category] = 0

        nb_user = 0
        labels = asset["labels"]
        for label in labels:
            if label["labelType"] == "DEFAULT" and label["isLatestLabelForUser"]:
                nb_user += 1
                response = json.loads(label["jsonResponse"])
                if "categories" not in response:
                    continue
                response_categories = response["categories"]
                for checked_category in response_categories:
                    dic_categories[checked_category["name"]] += 1
        consensus = 0
        if (nb_user - 1) == 0:
            continue
        for category in categories:
            consensus += (1.0 / (nb_user)) * dic_categories[category]
        consensus_by_asset[asset["id"]] = (1.0 / len(categories)) * consensus
    print(consensus_by_asset)
    return consensus_by_asset


def compute_consensus_for_project(client, project_id, interface_category, skip=0, first=100000):
    assets = get_assets(client, project_id, skip, first)
    if not assets:
        return None
    assets_for_consensus = []
    for asset in assets:
        if asset["isUsedForConsensus"] and (asset["status"] == "LABELED" or asset["status"] == "REVIEWED") and not \
                asset["isHoneypot"]:
            assets_for_consensus.append(asset)

    if interface_category == "IMAGE" or interface_category == "IMAGE_TO_TEXT":
        categories = list(json.loads(get_tools(client, project_id)[
            0]["jsonSettings"])["annotation_types"].keys())
        consensus_by_asset = {}
        for asset in assets_for_consensus:
            labels = asset["labels"]
            authors = compute_authors(labels)
            all_bounding_poly, categories = compute_bounding_polygons(
                labels, authors)
            kappa_matrices_by_category = compute_pixel_matrices_by_category(all_bounding_poly, categories, authors,
                                                                            grid_definition=100)

            kappa_mean_over_categories = 0
            for category in categories:
                kappa_mean_over_categories += fleiss_kappa(
                    kappa_matrices_by_category[category], method="fleiss")
                print("Asset: {}, Category: {}, Fleiss-Kappa: {}".format(asset["id"], category,
                                                                         fleiss_kappa(
                                                                             kappa_matrices_by_category[category],
                                                                             method="fleiss")))
            if len(categories) > 0:
                consensus_by_asset[asset["id"]] = max(
                    0, kappa_mean_over_categories / len(categories))
            else:
                consensus_by_asset[asset["id"]] = 0
        if interface_category == "IMAGE":
            print("Image consensus :", consensus_by_asset)
            return consensus_by_asset
        elif interface_category == "IMAGE_TO_TEXT":
            consensus_by_asset_texts = compute_consensus_for_project_ocr_texts(
                assets_for_consensus, categories)
            overall_consensus_by_asset = {}
            for id in consensus_by_asset.keys():
                overall_consensus_by_asset[id] = (
                    consensus_by_asset[id] + consensus_by_asset_texts[id]) / 2
            print("Image consensus :", consensus_by_asset)
            print("Overall consensus :", overall_consensus_by_asset)
            return overall_consensus_by_asset

    elif interface_category == "SINGLECLASS_TEXT_CLASSIFICATION" or interface_category == "MULTICLASS_TEXT_CLASSIFICATION" \
            or interface_category == "VIDEO_CLASSIFICATION":
        return compute_consensus_for_assets(assets_for_consensus)


def list_to_doubles(l):
    result = []
    for i in range(len(l)):
        for j in range(i):
            result.append((l[i], l[j]))
    return result


def compute_text_similarity_for_list_of_texts(texts, categories):
    lists = dict((cat, []) for cat in categories)
    for cat in categories:
        for text in texts:
            lists[cat].append(text[cat])
    for cat in categories:
        distances = [SequenceMatcher(None, a[0], a[1]).ratio()
                     for a in list_to_doubles(lists[cat])]
        lists[cat] = sum(distances) / \
            len(distances) if len(distances) > 0 else 0
    values = lists.values()
    return sum(values) / len(values) if len(values) > 0 else 0


def compute_consensus_for_project_ocr_texts(assets_for_consensus, categories):
    consensus_by_asset = {}
    texts_by_asset = dict.fromkeys([asset["id"]
                                    for asset in assets_for_consensus], [])
    for asset in assets_for_consensus:
        labels = [label for label in asset["labels"]
                  if label["isLatestLabelForUser"]]
        for label in labels:
            texts = dict.fromkeys(categories, "")
            annotations = json.loads(label["jsonResponse"])["annotations"]
            for annotation in annotations:
                category = list(annotation["description"][0].keys())[0]
                texts[category] = annotation["text"]
            texts_by_asset[asset["id"]].append(texts)

        consensus_by_asset[asset["id"]] = compute_text_similarity_for_list_of_texts(texts_by_asset[asset["id"]],
                                                                                    categories)
    print("Text consensus :", consensus_by_asset)
    return consensus_by_asset


def force_consensus_for_project(client, project_id: str):
    interface_category = get_project(client, project_id)['interfaceCategory']
    print(interface_category)
    consensus_by_asset = compute_consensus_for_project(
        client, project_id, interface_category)
    if not consensus_by_asset:
        return "WARNING : No asset in your project."
    asset_ids = list(consensus_by_asset.keys())
    consensus_marks = list(consensus_by_asset.values())
    are_used_for_consensus = [True for _ in consensus_marks]

    if len(asset_ids) > 0:
        update_consensus_in_many_assets(
            client, asset_ids, consensus_marks, are_used_for_consensus)



import os
import requests
import warnings

from .graphql_client import GraphQLClient

from .mutations.user import signin

warnings.simplefilter('always')
warnings.filterwarnings('ignore', category=ImportWarning)


MAX_RETRIES = 20


def authenticate(email,
                 password=os.getenv('KILI_USER_PASSWORD'),
                 api_endpoint='https://cloud.kili-technology.com/api/label/graphql'):

    message = '''Deprecated: authenticate has been renamed, and is not recommended for use.
Use instead:
  from kili.authentication import KiliAuth
  from kili.playground import Playground
  kauth = KiliAuth()
  playground = Playground(kauth)
  assets = playground.export_assets(project_id)'''

    warnings.warn(message, DeprecationWarning)

    session = requests.Session()
    adapter = requests.adapters.HTTPAdapter(max_retries=MAX_RETRIES)
    session.mount('https://', adapter)
    session.mount('http://', adapter)
    client = GraphQLClient(api_endpoint, session)

    auth_payload = signin(client, email, password)
    api_token = auth_payload['token']
    client.inject_token('Bearer: ' + api_token)
    user_id = auth_payload['user']['id']
    return client, user_id


class KiliAuth(object):
    """
    from kili.authentication import KiliAuth
    from kili.playground import Playground
    kauth = KiliAuth()
    playground = Playground(kauth)
    assets = playground.export_assets(project_id)
    """

    def __init__(self,
                 email=os.getenv('KILI_USER_EMAIL'),
                 password=os.getenv('KILI_USER_PASSWORD'),
                 api_endpoint='https://cloud.kili-technology.com/api/label/graphql'):
        self.session = requests.Session()
        adapter = requests.adapters.HTTPAdapter(max_retries=MAX_RETRIES)
        self.session.mount('https://', adapter)
        self.session.mount('http://', adapter)
        self.client = GraphQLClient(api_endpoint, self.session)
        auth_payload = signin(self.client, email, password)
        api_token = auth_payload['token']
        self.client.inject_token('Bearer: ' + api_token)
        self.user_id = auth_payload['user']['id']

    def __del__(self):
        self.session.close()



from json import dumps
from typing import List

from ..helper import content_escape, format_result, json_escape


def create_assets(client, project_id: str, contents: List[str], external_ids: List[str]):
    result = client.execute('''
    mutation {
      createAssets(
        projectID: "%s",
        contents: %s,
        externalIDs: %s) {
          id
          content
          externalId
          createdAt
          updatedAt
          isHoneypot
          status
          jsonMetadata
      }
    }
    ''' % (project_id, dumps(contents), dumps(external_ids)))
    return format_result('createAssets', result)


def delete_assets_by_external_id(client, project_id: str, external_id: str):
    result = client.execute('''
    mutation {
      deleteAssetsByExternalId(projectID: "%s", externalID: "%s") {
        id
      }
    }
    ''' % (project_id, external_id))
    return format_result('deleteAssetsByExternalId', result)


def append_to_dataset(client, project_id: str, content: str, external_id: str, filename: str = '', is_instructions: bool = False,
                      instructions: str = '', is_honeypot: bool = False, status: str = 'TODO', json_metadata: dict = {}):
    print("append_to_dataset: This method will be deprecated soon. Please use append_many_to_dataset instead.")
    return append_many_to_dataset(client, project_id,
                                  content_array=[content],
                                  external_id_array=[external_id],
                                  filename_array=[filename],
                                  is_instructions_array=[is_instructions],
                                  instructions_array=[instructions],
                                  is_honeypot_array=[is_honeypot],
                                  status_array=[status],
                                  json_metadata_array=[json_metadata])


def append_many_to_dataset(client, project_id: str, content_array: List[str], external_id_array: List[str],
                           filename_array: List[str] = None, is_instructions_array: List[bool] = None, instructions_array: List[str] = None,
                           is_honeypot_array: List[bool] = None, status_array: List[str] = None, json_metadata_array: List[str] = None):
    filename_array = [
        ''] * len(content_array) if not filename_array else filename_array
    is_instructions_array = [
        False] * len(content_array) if not is_instructions_array else is_instructions_array
    instructions_array = [
        ''] * len(content_array) if not instructions_array else instructions_array
    is_honeypot_array = [
        False] * len(content_array) if not is_honeypot_array else is_honeypot_array
    status_array = ['TODO'] * \
        len(content_array) if not status_array else status_array
    json_metadata_array = [
        {}] * len(content_array) if not json_metadata_array else json_metadata_array
    result = client.execute('''
        mutation {
          appendManyToDataset(
            projectID: "%s",
            contentArray: %s,
            externalIDArray: %s,
            filenameArray: %s,
            isInstructionsArray: %s,
            instructionsArray: %s,
            isHoneypotArray: %s,
            statusArray: %s,
            jsonMetadataArray: %s) {
            id
          }
        }
    ''' % (project_id, dumps(content_array), dumps(external_id_array), dumps(filename_array),
           dumps(is_instructions_array).lower(), dumps(
               instructions_array), dumps(is_honeypot_array).lower(),
           dumps(status_array).replace('"', ''), dumps([dumps(elem) for elem in json_metadata_array])))
    return format_result('appendManyToDataset', result)


def update_asset(client, asset_id: str, project_id: str, content: str, external_id: str, filename: str, is_instructions: bool, instructions: str,
                 is_honeypot: bool, consensus_mark: float, honeypot_mark: float, status: str, json_metadata: str):
    result = client.execute('''
    mutation {
      updateAsset(
        assetID: "%s",
        projectID: "%s",
        content: "%s",
        externalID: "%s",
        filename: "%s",
        isInstructions: %s,
        instructions: "%s",
        isHoneypot: %s,
        consensusMark: %d,
        honeypotMark: %d,
        status: %s,
        jsonMetadata: "%s") {
        id
      }
    }
    ''' % (asset_id, project_id, content, external_id, filename, str(is_instructions).lower(), instructions,
           str(is_honeypot).lower(), consensus_mark, honeypot_mark, status, json_escape(json_metadata)))
    return format_result('updateAsset', result)


def update_properties_in_asset(client, asset_id: str, external_id: str = None,
                               priority: int = None, json_metadata: str = None, consensus_mark: float = None,
                               honeypot_mark: float = None, to_be_labeled_by: List[str] = None):
    formatted_external_id = 'null' if external_id is None else f'"{external_id}"'
    formatted_priority = 'null' if priority is None else f'{priority}'
    formatted_json_metadata = 'null'
    if json_metadata is None:
        formatted_json_metadata = 'null'
    elif isinstance(json_metadata, str):
        formatted_json_metadata = f'"{json_metadata}"'
    elif isinstance(json_metadata, dict) or isinstance(json_metadata, list):
        formatted_json_metadata = f'"{json_escape(json_metadata)}"'
    else:
        raise Exception('json_metadata',
                        'Should be either a dict, a list or a string url')
    formatted_consensus_mark = 'null' if consensus_mark is None else f'{consensus_mark}'
    formatted_honeypot_mark = 'null' if honeypot_mark is None else f'{honeypot_mark}'
    formatted_to_be_labeled_by = 'null' if to_be_labeled_by is None else f'{dumps(to_be_labeled_by)}'

    result = client.execute('''
        mutation {
          updatePropertiesInAsset(
            where: {id: "%s"},
            data: {
              externalId: %s
              priority: %s
              jsonMetadata: %s
              consensusMark: %s
              honeypotMark: %s
              toBeLabeledBy: %s
            }
          ) {
            id
          }
        }
        ''' % (asset_id, formatted_external_id, formatted_priority,
               formatted_json_metadata, formatted_consensus_mark,
               formatted_honeypot_mark, formatted_to_be_labeled_by))
    return format_result('updatePropertiesInAsset', result)


def delete_from_dataset(client, asset_id: str):
    result = client.execute('''
    mutation {
      deleteFromDataset(assetID: "%s") {
        id
      }
    }
    ''' % (asset_id))
    return format_result('deleteFromDataset', result)


def delete_many_from_dataset(client, asset_ids: List[str]):
    result = client.execute('''
    mutation {
      deleteManyFromDataset(assetIDs: %s) {
        id
      }
    }
    ''' % (dumps(asset_ids)))
    return format_result('deleteManyFromDataset', result)


def force_update_status(client, asset_id: str):
    result = client.execute('''
    mutation {
      forceUpdateStatus(assetID: "%s") {
        id
        status
      }
    }
    ''' % (asset_id))
    return format_result('forceUpdateStatus', result)



import subprocess
import time

from kili.authentication import KiliAuth
from kili.playground import Playground

SECONDS_TO_WAIT = 10


def get_labels_of_types(asset, label_types):
    labels = [label for label in asset['labels']
              if label['labelType'] in label_types and label['isLatestLabelForUser']]
    return sorted(labels, key=lambda label: label['createdAt'])


class TransferLearning:
    def __init__(self, email, password, api_endpoint, project_id, minimum_number_of_assets_to_launch_training=100):
        kauth = KiliAuth(email, password, api_endpoint=api_endpoint)

        self.playground = Playground(kauth)
        self.project_id = project_id
        self.current_training_number = 0
        self.last_training_number = -1
        self.assets_seen_in_training = []
        self.minimum_number_of_assets_to_launch_training = minimum_number_of_assets_to_launch_training

    def _current_training_number(self):
        return self.current_training_number

    def get_assets_to_train(self):
        assets = self.playground.get_assets(project_id=self.project_id)
        assets_to_train = []
        for asset in assets:
            default_labels = get_labels_of_types(asset, ['DEFAULT'])
            review_labels = get_labels_of_types(asset, ['REVIEWED'])
            if len(review_labels) > 0:
                asset['labels'] = [review_labels[-1]]
                assets_to_train.append(asset)
            elif len(default_labels) == 1:
                asset['labels'] = [default_labels[-1]]
                assets_to_train.append(asset)
            elif len(review_labels) == 0 and len(default_labels) > 0:
                print(f'Asset {asset["id"]} has several labels: it should be reviewed')
            else:
                continue

        return assets_to_train

    def train(self, assets_to_train):
        print(f'Launch training for {len(assets_to_train)} assets: {[asset["id"] for asset in assets_to_train]}')
        return

    def launch_train(self):
        time.sleep(SECONDS_TO_WAIT)
        assets_to_train = self.get_assets_to_train()
        if len(self.assets_seen_in_training) == 0:
            filtered_assets_to_train = assets_to_train
        else:
            filtered_assets_to_train = [asset for asset in assets_to_train
                                        if all([asset['id'] not in training
                                                for training in self.assets_seen_in_training])]
        if len(filtered_assets_to_train) >= self.minimum_number_of_assets_to_launch_training:
            self.train(filtered_assets_to_train)
            self.current_training_number += 1
            self.assets_seen_in_training.append([asset['id'] for asset in filtered_assets_to_train])

    def get_assets_to_predict(self):
        assets = self.playground.get_assets(project_id=self.project_id)
        assets_to_predict = []
        for asset in assets:
            labels = get_labels_of_types(asset, ['DEFAULT', 'REVIEWED'])

            if len(labels) == 0:
                assets_to_predict.append(asset)

        return assets_to_predict

    def predict(self, assets_to_predict):
        print(f'Launch inference for {len(assets_to_predict)} assets: {[asset["id"] for asset in assets_to_predict]}')
        return

    def launch_predict(self):
        time.sleep(SECONDS_TO_WAIT)
        if self.current_training_number == 0 or self.current_training_number == self.last_training_number:
            print('Inference will not be launched for now...')
            return
        assets_to_predict = self.get_assets_to_predict()
        if len(assets_to_predict) > 0:
            current_training_number = self.current_training_number
            self.predict(assets_to_predict)
            self.last_training_number = current_training_number

    def launch_tensorboard(self):
        print('Starting Tensorboard...')
        subprocess.Popen(['tensorboard', '--logdir=runs'])
        print('You can access Tensorboard at http://localhost:6006\n')

    def launch(self):
        self.launch_tensorboard()
        while True:
            self.launch_train()
            self.launch_predict()



from ..helper import format_result


def get_projects(client, user_id: str):
    result = client.execute('''
    query {
      getProjects(userID: "%s") {
        id
        title
        consensusTotCoverage
        maxWorkerCount
        minAgreement
        minConsensusSize
        numberOfAssets
        numberOfRemainingAssets
        numberOfAssetsWithSkippedLabels
        numberOfReviewedAssets
        numberOfLatestLabels
        roles {
          id
          user { id, name, email }
          role
          lastLabelingAt
          numberOfAnnotations
          numberOfLabeledAssets
          totalDuration
          honeypotMark
        }
        dataset {
          id
          honeypotMark
        }
      }
    }
    ''' % (user_id))
    return format_result('getProjects', result)


def get_project(client, project_id: str):
    result = client.execute('''
    query {
      getProject(projectID: "%s") {
        id
        title
        numberOfAssets
        interfaceCategory
        calculatedConsensusMark
        calculatedHoneypotMark
        consensusMark
        roles {
          id
          user { id, name, email }
          role
          lastLabelingAt
          numberOfAnnotations
          numberOfLabeledAssets
          totalDuration
          honeypotMark
          calculatedConsensusMark
          calculatedHoneypotMark
          consensusMark
        }
      }
    }
    ''' % (project_id))
    return format_result('getProject', result)



import time
from json import dumps, loads
from typing import List

from tqdm import tqdm

from ..helper import GraphQLError, format_result, json_escape
from ..queries.asset import get_assets
from ..queries.label import get_label
from ..queries.project import get_project
from .asset import force_update_status, update_properties_in_asset
from .lock import delete_locks


def create_project(client, title: str, description: str, tool_type: str, use_honeypot: bool,
                   interface_json_settings: str):
    result = client.execute('''
    mutation {
      createProject(
      title: "%s",
      description: "%s",
      toolType: %s,
      useHoneyPot: %s,
      interfaceJsonSettings: "%s") {
        id
      }
    }
    ''' % (title, description, tool_type, str(use_honeypot).lower(), json_escape(interface_json_settings)))
    return format_result('createProject', result)


def delete_project(client, project_id: str):
    result = client.execute('''
    mutation {
      deleteProject(projectID: "%s") {
        id
      }
    }
    ''' % (project_id))
    return format_result('deleteProject', result)


def append_to_roles(client, project_id: str, user_email: str, role: str):
    result = client.execute('''
    mutation {
      appendToRoles(
        projectID: "%s",
        userEmail: "%s",
        role: %s) {
          id
          title
          roles {
              user {
                id
                email
              }
              role
         }
      }
    }
    ''' % (project_id, user_email, role))
    return format_result('appendToRoles', result)


def update_properties_in_project(client, project_id: str, min_consensus_size: int = None,
                                 consensus_tot_coverage: int = None,
                                 number_of_assets: int = None,
                                 number_of_remaining_assets: int = None,
                                 number_of_assets_with_empty_labels: int = None,
                                 number_of_reviewed_assets: int = None,
                                 number_of_latest_labels: int = None,
                                 consensus_mark: float = None,
                                 honeypot_mark: float = None,
                                 instructions: str = None):
    formatted_min_consensus_size = 'null' if min_consensus_size is None else int(
        min_consensus_size)
    formatted_consensus_tot_coverage = 'null' if consensus_tot_coverage is None else int(
        consensus_tot_coverage)
    formatted_number_of_assets = 'null' if number_of_assets is None else int(
        number_of_assets)
    formatted_number_of_remaining_assets = 'null' if number_of_remaining_assets is None else int(
        number_of_remaining_assets)
    formatted_number_of_assets_with_empty_labels = 'null' if number_of_assets_with_empty_labels is None else int(
        number_of_assets_with_empty_labels)
    formatted_number_of_reviewed_assets = 'null' if number_of_reviewed_assets is None else int(
        number_of_reviewed_assets)
    formatted_number_of_latest_labels = 'null' if number_of_latest_labels is None else int(
        number_of_latest_labels)
    formatted_consensus_mark = 'null' if consensus_mark is None else float(
        consensus_mark)
    formatted_honeypot_mark = 'null' if honeypot_mark is None else float(
        honeypot_mark)
    formatted_instructions = 'null' if instructions is None else f'{dumps(instructions)}'

    result = client.execute('''
        mutation {
          updatePropertiesInProject(
            where: {id: "%s"},
            data: {
              minConsensusSize: %s
              consensusTotCoverage: %s
              numberOfAssets: %s
              numberOfRemainingAssets: %s
              numberOfAssetsWithSkippedLabels: %s
              numberOfReviewedAssets: %s
              numberOfLatestLabels: %s
              consensusMark: %s
              honeypotMark: %s
              instructions: %s
            }
          ) {
            id
          }
        }
        ''' % (project_id, formatted_min_consensus_size, formatted_consensus_tot_coverage,
               formatted_number_of_assets, formatted_number_of_remaining_assets,
               formatted_number_of_assets_with_empty_labels, formatted_number_of_reviewed_assets,
               formatted_number_of_latest_labels, formatted_consensus_mark, formatted_honeypot_mark,
               formatted_instructions))
    return format_result('updatePropertiesInProject', result)


def update_interface_in_project(client, project_id: str, jsonSettings: str = None):
    result = client.execute('''
        mutation {
          updatePropertiesInProject(
            where: {id: "%s"},
            data: {
              jsonSettings: """%s""",
            }
          ) {
            id
          }
        }
        ''' % (project_id, jsonSettings))
    return format_result('updatePropertiesInProject', result)


def create_empty_project(client, user_id: str):
    result = client.execute('''
    mutation {
      createEmptyProject(userID: "%s") {
        id

      }
    }
    ''' % (user_id))
    return format_result('createEmptyProject', result)


def update_project(client, project_id: str,
                   title: str,
                   description: str,
                   interface_category: str,
                   creation_active_step: int = 0,
                   creation_completed: List[int] = [0, 1, 2, 3, 4, 5, 6],
                   creation_skipped: List[int] = [],
                   input_type: str = 'TEXT',
                   interface_title: str = None,
                   interface_description: str = None,
                   interface_url: str = None,
                   outsource: bool = False,
                   consensus_tot_coverage: int = 0,
                   min_consensus_size: int = 1,
                   max_worker_count: int = 4,
                   min_agreement: int = 66,
                   use_honey_pot: bool = False,
                   instructions: str = None,
                   model_title: str = "",
                   model_description: str = "",
                   model_url: str = ""):
    formatted_interface_title = 'null' if interface_title is None else f'"{interface_title}"'
    formatted_interface_description = 'null' if interface_description is None else f'"{interface_description}"'
    formatted_interface_url = 'null' if interface_url is None else f'"{interface_url}"'
    formatted_instructions = 'null' if instructions is None else f'"{instructions}"'
    result = client.execute('''
    mutation {
      updateProject(projectID: "%s",
        title: "%s",
        description: "%s",
        creationActiveStep: %d,
        creationCompleted: %s,
        creationSkipped: %s,
        interfaceCategory: %s,
        inputType: %s,
        interfaceTitle: %s,
        interfaceDescription: %s,
        interfaceUrl: %s,
        outsource: %s,
        consensusTotCoverage: %d,
        minConsensusSize: %d,
        maxWorkerCount: %d,
        minAgreement: %d,
        useHoneyPot: %s,
        instructions: %s,
        modelTitle: "%s",
        modelDescription: "%s",
        modelUrl: "%s") {
        id
      }
    }
    ''' % (
        project_id, title, description, creation_active_step, dumps(
            creation_completed),
        dumps(creation_skipped).lower(),
        interface_category, input_type, formatted_interface_title, formatted_interface_description,
        formatted_interface_url, str(
            outsource).lower(),
        consensus_tot_coverage, min_consensus_size, max_worker_count, min_agreement,
        str(use_honey_pot).lower(), formatted_instructions, model_title, model_description, model_url))
    return format_result('updateProject', result)


def update_role(client, role_id: str, project_id: str, user_id: str, role: str):
    result = client.execute('''
    mutation {
      updateRole(roleID: "%s",
        projectID: "%s",
        userID: "%s",
        role: %s) {
          id
      }
    }
    ''' % (role_id, project_id, user_id, role))
    return format_result('updateRole', result)


def delete_from_roles(client, role_id: str):
    result = client.execute('''
    mutation {
      deleteFromRoles(roleID: "%s") {
        id
      }
    }
    ''' % (role_id))
    return format_result('deleteFromRoles', result)


def update_properties_in_project_user(client, project_user_id: str,
                                      total_duration: int = None,
                                      number_of_labeled_assets: int = None,
                                      consensus_mark: float = None,
                                      honeypot_mark: float = None):
    args = [total_duration,
            number_of_labeled_assets, consensus_mark, honeypot_mark]
    formatted_args = ['null' if arg is None else f'{arg}' for arg in args]

    result = client.execute('''
        mutation {
          updatePropertiesInProjectUser(
            where: {id: "%s"},
            data: {
              totalDuration: %s
              numberOfLabeledAssets: %s
              consensusMark: %s
              honeypotMark: %s
            }
          ) {
            id
          }
        }
        ''' % (project_user_id, *formatted_args))
    return format_result('updatePropertiesInProjectUser', result)


def force_project_kpis(client, project_id: str):
    project = get_project(client, project_id=project_id)
    assets = get_assets(client, project_id=project_id)
    numbers_of_labeled_assets = {}
    number_of_latest_labels = 0
    for asset in tqdm(assets):
        asset_updated = force_update_status(client, asset['id'])
        asset['status'] = asset_updated['status']
        unique_asset_authors = list(
            set([label['author']['id'] for label in asset['labels']]))
        json_metadata = asset['jsonMetadata']
        if asset['jsonMetadata'] is not None and not asset['jsonMetadata'].startswith('http') and asset['jsonMetadata'] != '':
            json_metadata = loads(asset['jsonMetadata'])
        update_properties_in_asset(client, asset['id'], external_id=asset['externalId'], priority=asset['priority'],
                                   json_metadata=json_metadata, consensus_mark=asset[
                                       'calculatedConsensusMark'],
                                   honeypot_mark=asset['calculatedHoneypotMark'])

        for asset_author in unique_asset_authors:
            numbers_of_labeled_assets[asset_author] = 1 if asset_author not in numbers_of_labeled_assets else \
                numbers_of_labeled_assets[asset_author] + 1

        for label in asset['labels']:
            label_author = label['author']
            if label['isLatestLabelForUser']:
                number_of_latest_labels += 1
            get_label(client, asset_id=asset['id'], user_id=label_author)
        delete_locks(client, asset['id'])
        time.sleep(1)

    # Refresh project
    get_project(client, project_id=project_id)

    number_of_assets = len([a for a in assets if not a['isInstructions']])
    number_of_remaining_assets = len(
        [a for a in assets if a['status'] == 'TODO' or a['status'] == 'ONGOING'])

    update_properties_in_project(client, project_id, number_of_assets=number_of_assets,
                                 number_of_remaining_assets=number_of_remaining_assets,
                                 number_of_latest_labels=number_of_latest_labels,
                                 consensus_mark=project['calculatedConsensusMark'],
                                 honeypot_mark=project['calculatedHoneypotMark'])

    project_users = project['roles']
    for project_user in project_users:
        total_duration = project_user['totalDuration']
        user_id = project_user['user']['id']
        number_of_labeled_assets = numbers_of_labeled_assets[
            user_id] if user_id in numbers_of_labeled_assets else 0
        try:
            update_properties_in_project_user(client, project_user['id'], total_duration=total_duration,
                                              number_of_labeled_assets=number_of_labeled_assets,
                                              consensus_mark=project_user['calculatedConsensusMark'],
                                              honeypot_mark=project_user['calculatedHoneypotMark'])
        except GraphQLError as e:
            if 'is trying to access projectUser' in str(e):
                print(f'Could not update {user_id}')
            else:
                raise e



import pandas as pd

from ..helper import format_result
from .asset import get_assets
from .project import get_project


def get_label(client, asset_id: str, user_id: str):
    result = client.execute('''
    query {
      getLabel(assetID: "%s", userID: "%s") {
        id
        jsonResponse
        numberOfAnnotations
      }
    }
    ''' % (asset_id, user_id))
    return format_result('getLabel', result)


def get_latest_labels_for_user(client, project_id: str, user_id: str):
    result = client.execute('''
    query {
      getLatestLabelsForUser(projectID: "%s", userID: "%s") {
        id
        jsonResponse
      }
    }
    ''' % (project_id, user_id))
    return format_result('getLatestLabelsForUser', result)


def get_latest_labels(client, project_id: str, skip: int, first: int):
    result = client.execute('''
    query {
      getLatestLabels(projectID: "%s", skip: %d, first: %d) {
        id
        jsonResponse
      }
    }
    ''' % (project_id, skip, first))
    return format_result('getLatestLabels', result)


def parse_json_response_for_single_classification(json_response):
    categories = parse_json_response_for_multi_classification(json_response)
    if len(categories) == 0:
        return []

    return categories[0]


def parse_json_response_for_multi_classification(json_response):
    formatted_json_response = eval(json_response)
    if 'categories' not in formatted_json_response:
        return []
    categories = formatted_json_response['categories']
    return list(map(lambda category: category['name'], categories))


def parse_json_response(json_response, interface_category):
    if interface_category == 'SINGLECLASS_TEXT_CLASSIFICATION':
        return parse_json_response_for_single_classification(json_response)
    if interface_category == 'MULTICLASS_TEXT_CLASSIFICATION':
        return parse_json_response_for_multi_classification(json_response)

    return json_response


def export_labels_as_df(client, project_id: str):
    project = get_project(client, project_id)
    if 'interfaceCategory' not in project:
        return pd.DataFrame()

    interface_category = project['interfaceCategory']
    assets = get_assets(client, project_id=project_id)
    labels = [dict(label, **dict((f'asset__{key}', asset[key]) for key in asset))
              for asset in assets for label in asset['labels']]
    labels_df = pd.DataFrame(labels)
    labels_df['y'] = labels_df['jsonResponse'].apply(
        lambda json_response: parse_json_response(json_response, interface_category))
    return labels_df



from json import dumps, loads


class GraphQLError(Exception):
    def __init__(self, mutation, error):
        super().__init__(f'Mutation "{mutation}" failed with error: "{error}"')


def format_result(name, result):
    if 'errors' in result:
        raise GraphQLError(name, result['errors'])
    return result['data'][name]


def json_escape(dict):
    str = dumps(dict)
    return dumps(str)[1:-1]


def content_escape(content):
    return content.replace('\\', '\\\\').replace('\n', '\\n').replace('"', '\\"')



from ..helper import format_result


def get_locks(client, project_id: str):
    result = client.execute('''
    query {
      getLocks(projectID: "%s") {
        id
        author {
          email
        }
        createdAt
        lockType
      }
    }
    ''' % (project_id))
    return format_result('getLocks', result)



from ..helper import format_result


def delete_locks(client, asset_id: str):
    result = client.execute('''
    mutation {
      deleteLocks(assetID: "%s") {
        id
      }
    }
    ''' % (asset_id))
    return format_result('deleteLocks', result)



from ..helper import format_result


def get_user(client, email: str):
    result = client.execute('''
    query {
      getUser(email: "%s") {
        id
      }
    }
    ''' % (email))
    return format_result('getUser', result)
